from dataclasses import dataclass
from typing import Any, Callable, Sequence, Type, TypeVar

from ..errors import CompositionError
from ..operation import Operation, get_operation
from ..services import Service
from ..specification import ClientSpecification
from ..typing import Dependency

__all__: Sequence[str] = ("response",)

CS = TypeVar("CS", Callable[..., Any], Type[Service])


@dataclass
class ResponseDecorator:
    dependency: Dependency

    def __call__(self, target: CS, /) -> CS:
        if isinstance(target, type):
            if not issubclass(target, Service):
                raise CompositionError(f"Target class is not a subclass of {Service}")

            client_specification: ClientSpecification = target._spec

            client_specification.default_response = self.dependency
        elif callable(target):
            operation: Operation = get_operation(target)

            operation.response = self.dependency
        else:
            raise CompositionError(f"Target of unsupported type {type(target)}")

        return target


response: Type[ResponseDecorator] = ResponseDecorator



from typing import Optional, Sequence

from httpx import URL, Limits, Timeout

from .types import (
    AuthTypes,
    CookiesTypes,
    DefaultEncodingTypes,
    EventHooks,
    HeadersTypes,
    QueriesTypes,
    TimeoutTypes,
    URLTypes,
)

__all__: Sequence[str] = (
    "DEFAULT_BASE_URL",
    "DEFAULT_AUTH",
    "DEFAULT_PARAMS",
    "DEFAULT_HEADERS",
    "DEFAULT_COOKIES",
    "DEFAULT_TIMEOUT",
    "DEFAULT_FOLLOW_REDIRECTS",
    "DEFAULT_MAX_REDIRECTS",
    "DEFAULT_EVENT_HOOKS",
    "DEFAULT_TRUST_ENV",
    "DEFAULT_ENCODING",
    "DEFAULT_LIMITS",
)

DEFAULT_BASE_URL: URLTypes = URL()
DEFAULT_AUTH: Optional[AuthTypes] = None
DEFAULT_PARAMS: Optional[QueriesTypes] = None
DEFAULT_HEADERS: Optional[HeadersTypes] = None
DEFAULT_COOKIES: Optional[CookiesTypes] = None
DEFAULT_TIMEOUT: TimeoutTypes = Timeout(timeout=5.0)
DEFAULT_FOLLOW_REDIRECTS: bool = False
DEFAULT_MAX_REDIRECTS: int = 20
DEFAULT_TRUST_ENV: bool = True
DEFAULT_ENCODING: DefaultEncodingTypes = "utf-8"
DEFAULT_LIMITS = Limits(max_connections=100, max_keepalive_connections=20)
DEFAULT_EVENT_HOOKS: Optional[EventHooks] = None



from dataclasses import dataclass
from typing import Callable, Optional, Sequence, Type, TypeVar

from httpx import URL
from mediate.protocols import MiddlewareCallable, MiddlewareMethod

from ..annotations import (
    service_middleware,
    service_request_dependency,
    service_response,
    service_response_dependency,
)
from ..models import Request, Response
from ..services import Service
from ..typing import Decorator, Dependency

__all__: Sequence[str] = ("service",)

C = TypeVar("C", bound=Callable)
M = TypeVar(
    "M", MiddlewareCallable[Request, Response], MiddlewareMethod[Request, Response]
)
S = TypeVar("S", bound=Type[Service])


@dataclass
class ServiceDecorator(Decorator[S]):
    base_url: Optional[str] = None
    middlewares: Optional[Sequence[MiddlewareCallable[Request, Response]]] = None
    default_response: Optional[Dependency] = None
    request_dependencies: Optional[Sequence[Dependency]] = None
    response_dependencies: Optional[Sequence[Dependency]] = None

    def __init__(
        self,
        base_url: Optional[str] = None,
        *,
        middleware: Optional[Sequence[MiddlewareCallable[Request, Response]]] = None,
        default_response: Optional[Dependency] = None,
        request_dependencies: Optional[Sequence[Dependency]] = None,
        response_dependencies: Optional[Sequence[Dependency]] = None,
    ) -> None:
        self.base_url = base_url
        self.middlewares = middleware
        self.default_response = default_response
        self.request_dependencies = request_dependencies
        self.response_dependencies = response_dependencies

    def __call__(self, target: S, /) -> S:
        if self.base_url is not None:
            target._spec.options.base_url = URL(self.base_url)
        if self.middlewares is not None:
            target._spec.middleware.add_all(self.middlewares)
        if self.default_response is not None:
            target._spec.default_response = self.default_response
        if self.request_dependencies is not None:
            target._spec.request_dependencies.extend(self.request_dependencies)
        if self.response_dependencies is not None:
            target._spec.response_dependencies.extend(self.response_dependencies)

        return target

    @staticmethod
    def middleware(middleware: M, /) -> M:
        return service_middleware(middleware)

    @staticmethod
    def response(response: C, /) -> C:
        return service_response(response)

    @staticmethod
    def request_depends(dependency: C, /) -> C:
        return service_request_dependency(dependency)

    @staticmethod
    def response_depends(dependency: C, /) -> C:
        return service_response_dependency(dependency)


service: Type[ServiceDecorator] = ServiceDecorator



from dataclasses import dataclass
from typing import Callable, Optional, Sequence, TypeVar

from typing_extensions import ParamSpec

from ..client import Client
from ..enums import HTTPMethod
from ..typing import Dependency

__all__: Sequence[str] = (
    "request",
    "put",
    "get",
    "post",
    "head",
    "patch",
    "delete",
    "options",
)

PS = ParamSpec("PS")
RT = TypeVar("RT")


def request(
    method: str,
    endpoint: str,
    /,
    *,
    response: Optional[Dependency] = None,
) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
    return Client().request(method, endpoint, response=response)


@dataclass
class MethodRequestDecorator:
    method: str

    def __call__(
        self,
        endpoint: str,
        /,
        *,
        response: Optional[Dependency] = None,
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return request(self.method, endpoint, response=response)


put: MethodRequestDecorator = MethodRequestDecorator(HTTPMethod.PUT)
get: MethodRequestDecorator = MethodRequestDecorator(HTTPMethod.GET)
post: MethodRequestDecorator = MethodRequestDecorator(HTTPMethod.POST)
head: MethodRequestDecorator = MethodRequestDecorator(HTTPMethod.HEAD)
patch: MethodRequestDecorator = MethodRequestDecorator(HTTPMethod.PATCH)
delete: MethodRequestDecorator = MethodRequestDecorator(HTTPMethod.DELETE)
options: MethodRequestDecorator = MethodRequestDecorator(HTTPMethod.OPTIONS)



from typing import Any, Callable, Sequence, TypeVar

from ..consumers import (
    ContentConsumer,
    DataConsumer,
    FilesConsumer,
    JsonConsumer,
    MountConsumer,
    PathConsumer,
    PathsConsumer,
)
from ..types import (
    JsonTypes,
    PathsTypes,
    PathTypes,
    RequestContent,
    RequestData,
    RequestFiles,
)
from .api import OperationConsumerDecorator

__all__: Sequence[str] = (
    "content",
    "data",
    "files",
    "json",
    "mount",
    "path",
    "path_params",
)

C = TypeVar("C", bound=Callable[..., Any])


def content(content: RequestContent, /) -> Callable[[C], C]:
    return OperationConsumerDecorator(ContentConsumer(content))


def data(data: RequestData, /) -> Callable[[C], C]:
    return OperationConsumerDecorator(DataConsumer(data))


def files(files: RequestFiles, /) -> Callable[[C], C]:
    return OperationConsumerDecorator(FilesConsumer(files))


def json(json: JsonTypes, /) -> Callable[[C], C]:
    return OperationConsumerDecorator(JsonConsumer(json))


def mount(path: str, /) -> Callable[[C], C]:
    return OperationConsumerDecorator(MountConsumer(path))


def path(key: str, value: PathTypes) -> Callable[[C], C]:
    return OperationConsumerDecorator(PathConsumer(key, value))


def path_params(path_params: PathsTypes, /) -> Callable[[C], C]:
    return OperationConsumerDecorator(PathsConsumer(path_params))



r"""
```
                       _ _            _   
 _ __   ___  ___   ___| (_) ___ _ __ | |_ 
| '_ \ / _ \/ _ \ / __| | |/ _ \ '_ \| __|
| | | |  __/ (_) | (__| | |  __/ | | | |_ 
|_| |_|\___|\___/ \___|_|_|\___|_| |_|\__|

       Fast API Clients for Python
       ~~~~~~~~~~~~~~~~~~~~~~~~~~~
               @tombulled
```
"""

__version__: str = "0.1.38"

from .client import NeoClient
from .decorators import (
    accept,
    base_url,
    content,
    cookie,
    cookies,
    data,
    files,
    header,
    headers,
    json,
    mount,
    path,
    path_params,
    query,
    query_params,
    referer,
    request_depends,
    response,
    response_depends,
    service,
    timeout,
    user_agent,
    verify,
)
from .decorators.middleware import (
    expect_content_type,
    expect_header,
    expect_status,
    middleware,
    raise_for_status,
)
from .decorators.request import delete, get, head, options, patch, post, put, request
from .models import Request, Response
from .param_functions import (
    URL,
    Body,
    Cookie,
    Cookies,
    Depends,
    Header,
    Headers,
    Path,
    Paths,
    Queries,
    Query,
    Reason,
    State,
    StatusCode,
)
from .services import Service



from typing import Any, Callable, Sequence, Type, TypeVar

from neoclient.services import Service

from ..param_functions import Request
from .common import request_depends

__all__: Sequence[str] = ("persist_pre_request",)

CS = TypeVar("CS", Callable[..., Any], Type[Service])


def _persist_pre_request_dependency(request=Request()):
    # TOOO: `request` is currently un-typed due to bug #136
    request.state.pre_request = request


def persist_pre_request(target: CS, /) -> CS:
    return request_depends(_persist_pre_request_dependency)(target)



from typing import Sequence

from annotate import Annotation

from .enums import Entity

__all__: Sequence[str] = (
    "service_middleware",
    "service_response",
    "service_request_dependency",
    "service_response_dependency",
)

service_middleware: Annotation = Annotation(Entity.MIDDLEWARE)
service_response: Annotation = Annotation(Entity.RESPONSE)
service_request_dependency: Annotation = Annotation(Entity.REQUEST_DEPENDENCY)
service_response_dependency: Annotation = Annotation(Entity.RESPONSE_DEPENDENCY)



import inspect
import string
import typing
from typing import (
    Any,
    Callable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
)

from pydantic import BaseConfig, BaseModel, create_model
from pydantic.fields import FieldInfo, Undefined
from pydantic.typing import display_as_type

__all__: Sequence[str] = (
    "parse_format_string",
    "bind_arguments",
    "is_primitive",
    "unpack_arguments",
    "get_default",
    "has_default",
    "parse_obj_as",
    "is_generic_alias",
)

T = TypeVar("T")


def parse_format_string(format_string: str, /) -> Set[str]:
    """
    Extracts a set of field names from `format_string`

    Example:
        >>> parse_format_string("foo {bar}")
        {"bar"}
    """

    path_params: Set[str] = set()

    field_name: Optional[str]
    for _, field_name, _, _ in string.Formatter().parse(format_string):
        if field_name is None:
            continue

        if not field_name.isidentifier():
            raise ValueError(f"Field name {field_name!r} is not a valid identifier")

        path_params.add(field_name)

    return path_params


def bind_arguments(
    func: Callable, /, args: Tuple[Any, ...], kwargs: Mapping[str, Any]
) -> Mapping[str, Any]:
    bound_arguments: inspect.BoundArguments = inspect.signature(func).bind(
        *args, **kwargs
    )

    bound_arguments.apply_defaults()

    return bound_arguments.arguments


def is_primitive(obj: Any, /) -> bool:
    return isinstance(obj, (str, int, float, bool, type(None)))


def unpack_arguments(
    func: Callable, arguments: Mapping[str, Any]
) -> Tuple[Tuple[Any, ...], Mapping[str, Any]]:
    parameters: Mapping[str, inspect.Parameter] = inspect.signature(func).parameters

    args: MutableSequence[Any] = []
    kwargs: MutableMapping[str, Any] = {}

    parameter: inspect.Parameter
    for parameter in parameters.values():
        if parameter.name not in arguments:
            raise ValueError(f"Missing argument for parameter {parameter.name!r}")

        argument: Any = arguments[parameter.name]

        if parameter.kind == inspect.Parameter.POSITIONAL_ONLY:
            args.append(argument)
        elif parameter.kind in (
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        ):
            kwargs[parameter.name] = argument
        elif parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            args.extend(argument)
        else:
            assert parameter.kind == inspect.Parameter.VAR_KEYWORD

            kwargs.update(argument)

    return (tuple(args), kwargs)


def get_default(field_info: FieldInfo, /) -> Any:
    if field_info.default_factory is not None:
        return field_info.default_factory()

    return field_info.default


def has_default(field_info: FieldInfo, /) -> bool:
    return field_info.default is not Undefined or field_info.default_factory is not None


def parse_obj_as(type_: Type[T], obj: Any) -> T:
    parsing_type_name: str = f"ParsingModel[{display_as_type(type_)}]"

    class Config(BaseConfig):
        arbitrary_types_allowed: bool = True

    model_cls: Type[BaseModel] = create_model(
        parsing_type_name,
        __config__=Config,
        __root__=(type_, ...),
    )

    model: BaseModel = model_cls(__root__=obj)

    return getattr(model, "__root__")


def is_generic_alias(type_: Type, /) -> bool:
    return typing.get_origin(type_) is not None



from datetime import timedelta
from typing import Any, Dict, List, Optional, Sequence

from httpx import URL, Cookies, Headers, QueryParams, Request, Response

from .types import StreamTypes

__all__: Sequence[str] = (
    "charset_encoding",
    "content",
    "elapsed",
    "encoding",
    "has_redirect_location",
    "headers",
    "history",
    "http_version",
    "is_client_error",
    "is_closed",
    "is_error",
    "is_informational",
    "is_redirect",
    "is_server_error",
    "is_stream_consumed",
    "is_success",
    "json",
    "links",
    "next_request",
    "num_bytes_downloaded",
    "reason_phrase",
    "request",
    "response",
    "status_code",
    "stream",
    "text",
    "url",
    "request_content",
    "request_headers",
    "request_method",
    "request_params",
    "request_stream",
    "request_url",
)


def charset_encoding(response: Response) -> Optional[str]:
    return response.charset_encoding


def content(response: Response) -> bytes:
    return response.content


def cookies(response: Response) -> Cookies:
    return response.cookies


def elapsed(response: Response) -> timedelta:
    return response.elapsed


def encoding(response: Response) -> Optional[str]:
    return response.encoding


def has_redirect_location(response: Response) -> bool:
    return response.has_redirect_location


def headers(response: Response) -> Headers:
    return response.headers


def history(response: Response) -> List[Response]:
    return response.history


def http_version(response: Response) -> str:
    return response.http_version


def is_client_error(response: Response) -> bool:
    return response.is_client_error


def is_closed(response: Response) -> bool:
    return response.is_closed


def is_error(response: Response) -> bool:
    return response.is_error


def is_informational(response: Response) -> bool:
    return response.is_informational


def is_redirect(response: Response) -> bool:
    return response.is_redirect


def is_server_error(response: Response) -> bool:
    return response.is_server_error


def is_stream_consumed(response: Response) -> bool:
    return response.is_stream_consumed


def is_success(response: Response) -> bool:
    return response.is_success


def json(response: Response) -> Any:
    return response.json()


def links(response: Response) -> Dict[Optional[str], Dict[str, str]]:
    return response.links


def next_request(response: Response) -> Optional[Request]:
    return response.next_request


def num_bytes_downloaded(response: Response) -> int:
    return response.num_bytes_downloaded


def reason_phrase(response: Response) -> str:
    return response.reason_phrase


def request(response: Response) -> Request:
    return response.request


def response(response: Response) -> Response:
    return response


def status_code(response: Response) -> int:
    return response.status_code


def stream(response: Response) -> StreamTypes:
    return response.stream


def text(response: Response) -> str:
    return response.text


def url(response: Response) -> URL:
    return response.url


def request_content(request: Request) -> bytes:
    return request.content


def request_headers(request: Request) -> Headers:
    return request.headers


def request_method(request: Request) -> str:
    return request.method


def request_params(request: Request) -> QueryParams:
    return request.url.params


def request_stream(request: Request) -> StreamTypes:
    return request.stream


def request_url(request: Request) -> URL:
    return request.url



import urllib.parse
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Union,
)

import httpx
from httpx import URL, BaseTransport, Cookies, Headers, Limits, QueryParams, Timeout
from httpx._config import DEFAULT_MAX_REDIRECTS, DEFAULT_TIMEOUT_CONFIG

from . import converters, utils
from .constants import USER_AGENT
from .defaults import (
    DEFAULT_BASE_URL,
    DEFAULT_ENCODING,
    DEFAULT_FOLLOW_REDIRECTS,
    DEFAULT_LIMITS,
    DEFAULT_TRUST_ENV,
)
from .enums import HTTPHeader
from .errors import IncompatiblePathParameters, MissingStateError
from .types import (
    AsyncByteStream,
    AuthTypes,
    CertTypes,
    CookiesTypes,
    DefaultEncodingTypes,
    EventHooks,
    HeadersTypes,
    JsonTypes,
    MethodTypes,
    PathsTypes,
    ProxiesTypes,
    QueriesTypes,
    RequestContent,
    RequestData,
    RequestExtensions,
    RequestFiles,
    ResponseContent,
    ResponseExtensions,
    SyncByteStream,
    TimeoutTypes,
    URLTypes,
    VerifyTypes,
)

__all__: Sequence[str] = (
    # httpx models
    "Cookies",
    "Headers",
    "QueryParams",
    "URL",
    # neoclient models
    "State",
    "ClientOptions",
    "Request",
    "Response",
    "PreRequest",
)


class State:
    _state: MutableMapping[str, Any]

    def __init__(self, state: Optional[MutableMapping[str, Any]] = None):
        if state is None:
            state = {}

        super().__setattr__("_state", state)

    def __repr__(self) -> str:
        context: str = ", ".join(
            f"{key}={value!r}" for key, value in self._state.items()
        )
        return f"{type(self).__name__}({context})"

    def __eq__(self, value: object) -> bool:
        if isinstance(value, type(self)):
            return self._state == value._state

        return False

    def _set(self, key: str, value: Any) -> None:
        self._state[key] = value

    def _get(self, key: str) -> Any:
        if key in self._state:
            return self._state[key]

        raise MissingStateError(key=key)

    def _del(self, key: str) -> None:
        del self._state[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._set(key, value)

    def __setattr__(self, key: str, value: Any) -> None:
        self._set(key, value)

    def __getitem__(self, key: str) -> Any:
        try:
            return self._get(key)
        except MissingStateError as missing_state_error:
            raise KeyError(key) from missing_state_error

    def __getattr__(self, key: str) -> Any:
        try:
            return self._get(key)
        except MissingStateError as missing_state_error:
            raise AttributeError(
                f"{type(self).__name__!r} object has no attribute {key!r}"
            ) from missing_state_error

    def __delitem__(self, key: str) -> None:
        self._del(key)

    def __delattr__(self, key: str) -> None:
        self._del(key)

    def get(self, key: str, default: Any = None) -> Any:
        return self._state.get(key, default)


class Request(httpx.Request):
    state: State

    def __init__(
        self,
        method: MethodTypes,
        url: URLTypes,
        *,
        params: Optional[QueriesTypes] = None,
        headers: Optional[HeadersTypes] = None,
        cookies: Optional[CookiesTypes] = None,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        stream: Union[SyncByteStream, AsyncByteStream, None] = None,
        extensions: Optional[RequestExtensions] = None,
        state: Optional[State] = None,
    ):
        super().__init__(
            method=method,
            url=url,
            params=params,
            headers=headers,
            cookies=(
                converters.convert_cookies(cookies) if cookies is not None else None
            ),
            content=content,
            data=data,
            files=files,
            json=json,
            stream=stream,
            extensions=extensions,
        )

        if state is not None:
            self.state = state
        else:
            self.state = State()

    @classmethod
    def from_httpx_request(cls, httpx_request: httpx.Request, /) -> "Request":
        if hasattr(httpx_request, "_content"):
            request: Request = cls(
                method=httpx_request.method,
                url=httpx_request.url,
                headers=httpx_request.headers,
                extensions=httpx_request.extensions,
                stream=httpx_request.stream,
            )

            request._content = httpx_request.content

            return request

        return cls(
            method=httpx_request.method,
            url=httpx_request.url,
            headers=httpx_request.headers,
            extensions=httpx_request.extensions,
            stream=httpx_request.stream,
        )


class Response(httpx.Response):
    state: State

    def __init__(
        self,
        status_code: int,
        *,
        headers: Optional[HeadersTypes] = None,
        content: Optional[ResponseContent] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Any = None,
        stream: Union[SyncByteStream, AsyncByteStream, None] = None,
        request: Optional[httpx.Request] = None,
        extensions: Optional[ResponseExtensions] = None,
        history: Optional[List[httpx.Response]] = None,
        default_encoding: Union[str, Callable[[bytes], str]] = "utf-8",
        state: Optional[State] = None,
    ):
        super().__init__(
            status_code=status_code,
            headers=headers,
            content=content,
            text=text,
            html=html,
            json=json,
            stream=stream,
            request=request,
            extensions=extensions,
            history=history,
            default_encoding=default_encoding,
        )

        if state is not None:
            self.state = state
        else:
            self.state = State()

    @classmethod
    def from_httpx_response(cls, httpx_response: httpx.Response, /) -> "Response":
        if hasattr(httpx_response, "_content"):
            response: Response = cls(
                status_code=httpx_response.status_code,
                headers=httpx_response.headers,
                request=httpx_response.request,
                stream=httpx_response.stream,
            )

            response._content = httpx_response.content

            return response

        return cls(
            status_code=httpx_response.status_code,
            headers=httpx_response.headers,
            request=httpx_response.request,
            stream=httpx_response.stream,
        )


@dataclass(init=False)
class ClientOptions:
    auth: Optional[AuthTypes]
    params: QueryParams
    headers: Headers
    cookies: Cookies
    verify: VerifyTypes
    cert: Optional[CertTypes]
    http1: bool
    http2: bool
    proxies: Optional[ProxiesTypes]
    mounts: Mapping[str, BaseTransport]
    timeout: Timeout
    follow_redirects: bool
    limits: Limits
    max_redirects: int
    event_hooks: EventHooks
    base_url: URL
    transport: Optional[BaseTransport]
    app: Optional[Callable[..., Any]]
    trust_env: bool
    default_encoding: DefaultEncodingTypes

    def __init__(
        self,
        auth: Optional[AuthTypes] = None,
        params: Optional[QueriesTypes] = None,
        headers: Optional[HeadersTypes] = None,
        cookies: Optional[CookiesTypes] = None,
        verify: VerifyTypes = True,
        cert: Optional[CertTypes] = None,
        http1: bool = True,
        http2: bool = False,
        proxies: Optional[ProxiesTypes] = None,
        mounts: Optional[Mapping[str, BaseTransport]] = None,
        timeout: TimeoutTypes = DEFAULT_TIMEOUT_CONFIG,
        follow_redirects: bool = DEFAULT_FOLLOW_REDIRECTS,
        limits: Limits = DEFAULT_LIMITS,
        max_redirects: int = DEFAULT_MAX_REDIRECTS,
        event_hooks: Optional[EventHooks] = None,
        base_url: URLTypes = DEFAULT_BASE_URL,
        transport: Optional[BaseTransport] = None,
        app: Optional[Callable[..., Any]] = None,
        trust_env: bool = DEFAULT_TRUST_ENV,
        default_encoding: DefaultEncodingTypes = DEFAULT_ENCODING,
    ) -> None:
        self.auth = auth
        self.params = (
            converters.convert_query_params(params)
            if params is not None
            else QueryParams()
        )
        self.headers = (
            converters.convert_headers(headers) if headers is not None else Headers()
        )
        self.cookies = (
            converters.convert_cookies(cookies) if cookies is not None else Cookies()
        )
        self.verify = verify
        self.cert = cert
        self.http1 = http1
        self.http2 = http2
        self.proxies = proxies
        self.mounts = mounts if mounts is not None else {}
        self.timeout = (
            converters.convert_timeout(timeout) if timeout is not None else Timeout()
        )
        self.follow_redirects = follow_redirects
        self.limits = limits
        self.max_redirects = max_redirects
        self.event_hooks = (
            event_hooks if event_hooks is not None else {"request": [], "response": []}
        )
        self.base_url = URL(base_url)
        self.transport = transport
        self.app = app
        self.trust_env = trust_env
        self.default_encoding = default_encoding

    def build(self) -> httpx.Client:
        headers: Headers = Headers(self.headers)

        # Set a default User-Agent header
        headers.setdefault(HTTPHeader.USER_AGENT, USER_AGENT)

        return httpx.Client(
            auth=self.auth,
            params=self.params,
            headers=headers,
            cookies=self.cookies,
            verify=self.verify,
            cert=self.cert,
            http1=self.http1,
            http2=self.http2,
            proxies=self.proxies,
            mounts=self.mounts,
            timeout=self.timeout,
            follow_redirects=self.follow_redirects,
            limits=self.limits,
            max_redirects=self.max_redirects,
            event_hooks=self.event_hooks,
            base_url=self.base_url,
            transport=self.transport,
            app=self.app,
            trust_env=self.trust_env,
            default_encoding=self.default_encoding,
        )


@dataclass(init=False)
class PreRequest:
    method: str
    url: URL
    params: QueryParams
    headers: Headers
    cookies: Cookies
    content: Optional[RequestContent]
    data: Optional[RequestData]
    files: Optional[RequestFiles]
    json: Optional[JsonTypes]
    timeout: Optional[Timeout]
    path_params: MutableMapping[str, str]
    state: State

    def __init__(
        self,
        method: MethodTypes,
        url: URLTypes,
        *,
        params: Optional[QueriesTypes] = None,
        headers: Optional[HeadersTypes] = None,
        cookies: Optional[CookiesTypes] = None,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[JsonTypes] = None,
        timeout: Optional[TimeoutTypes] = None,
        path_params: Optional[PathsTypes] = None,
        state: Optional[State] = None,
    ) -> None:
        self.method = method if isinstance(method, str) else method.decode()
        self.url = URL(url)
        self.params = (
            converters.convert_query_params(params)
            if params is not None
            else QueryParams()
        )
        self.headers = (
            converters.convert_headers(headers) if headers is not None else Headers()
        )
        self.cookies = (
            converters.convert_cookies(cookies) if cookies is not None else Cookies()
        )
        self.content = content
        self.data = data
        self.files = files
        self.json = json
        self.timeout = (
            converters.convert_timeout(timeout) if timeout is not None else None
        )
        self.path_params = (
            converters.convert_path_params(path_params)
            if path_params is not None
            else {}
        )
        self.state = state if state is not None else State()

    def build_request(self, client: Optional[httpx.Client]) -> Request:
        url: str = self._get_formatted_url()

        extensions: Dict[str, Any] = {}

        if client is None:
            if self.timeout is not None:
                extensions["timeout"] = self.timeout.as_dict()

            return Request(
                self.method,
                url,
                params=self.params,
                headers=self.headers,
                cookies=self.cookies,
                content=self.content,
                data=self.data,
                files=self.files,
                json=self.json,
                extensions=extensions,
                state=self.state,
            )

        httpx_request: httpx.Request = client.build_request(
            self.method,
            url,
            params=self.params,
            headers=self.headers,
            cookies=self.cookies,
            content=self.content,
            data=self.data,
            files=self.files,
            json=self.json,
            timeout=self.timeout,
            extensions=extensions,
        )

        request: Request = Request.from_httpx_request(httpx_request)

        request.state = self.state

        return request

    def merge(self, pre_request: "PreRequest", /) -> "PreRequest":
        return self.__class__(
            method=pre_request.method,
            url=pre_request.url,
            params=self.params.merge(pre_request.params),
            headers=httpx.Headers({**self.headers, **pre_request.headers}),
            cookies=httpx.Cookies({**self.cookies, **pre_request.cookies}),
            content=(
                pre_request.content if pre_request.content is not None else self.content
            ),
            data=(pre_request.data if pre_request.data is not None else self.data),
            files=(pre_request.files if pre_request.files is not None else self.files),
            json=(pre_request.json if pre_request.json is not None else self.json),
            timeout=(
                pre_request.timeout if pre_request.timeout is not None else self.timeout
            ),
            path_params={
                **self.path_params,
                **pre_request.path_params,
            },
        )

    def clone(self) -> "PreRequest":
        return self.merge(PreRequest(method=self.method, url=self.url))

    def _get_formatted_url(self) -> str:
        return urllib.parse.unquote(str(self.url)).format(**self.path_params)

    def validate(self):
        url: str = urllib.parse.unquote(str(self.url))

        expected_path_params: Set[str] = utils.parse_format_string(url)
        actual_path_params: Set[str] = set(self.path_params.keys())

        # Validate path params are correct
        if expected_path_params != actual_path_params:
            raise IncompatiblePathParameters(
                f"Expected {tuple(expected_path_params)}, got {tuple(actual_path_params)}"
            )



from typing import Any, Callable, Optional, Sequence

from pydantic import Required

from .dependence import DependencyParameter
from .params import (
    BodyParameter,
    CookieParameter,
    CookiesParameter,
    HeaderParameter,
    HeadersParameter,
    PathParameter,
    PathsParameter,
    QueriesParameter,
    QueryParameter,
    ReasonParameter,
    RequestParameter,
    ResponseParameter,
    StateParameter,
    StatusCodeParameter,
    URLParameter,
)
from .typing import Supplier

__all__: Sequence[str] = (
    "Query",
    "Header",
    "Cookie",
    "Path",
    "Queries",
    "Headers",
    "Cookies",
    "Paths",
    "Body",
    "Depends",
    "URL",
    "Reason",
    "Request",
    "Response",
    "StatusCode",
)


def Query(
    name: Optional[str] = None,
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> QueryParameter:
    return QueryParameter(
        default=default,
        default_factory=default_factory,
        alias=name,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        allow_inf_nan=allow_inf_nan,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_length=min_length,
        max_length=max_length,
        regex=regex,
    )


def Header(
    name: Optional[str] = None,
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    convert_underscores: bool = True,
) -> HeaderParameter:
    return HeaderParameter(
        default=default,
        default_factory=default_factory,
        alias=name,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        allow_inf_nan=allow_inf_nan,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_length=min_length,
        max_length=max_length,
        regex=regex,
        convert_underscores=convert_underscores,
    )


def Cookie(
    name: Optional[str] = None,
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> CookieParameter:
    return CookieParameter(
        default=default,
        default_factory=default_factory,
        alias=name,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        allow_inf_nan=allow_inf_nan,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_length=min_length,
        max_length=max_length,
        regex=regex,
    )


def Path(
    name: Optional[str] = None,
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> PathParameter:
    return PathParameter(
        default=default,
        default_factory=default_factory,
        alias=name,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        allow_inf_nan=allow_inf_nan,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_length=min_length,
        max_length=max_length,
        regex=regex,
    )


def Queries(
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> QueriesParameter:
    return QueriesParameter(
        default=default,
        default_factory=default_factory,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
    )


def Headers(
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> HeadersParameter:
    return HeadersParameter(
        default=default,
        default_factory=default_factory,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
    )


def Cookies(
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> CookiesParameter:
    return CookiesParameter(
        default=default,
        default_factory=default_factory,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
    )


def Paths(
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> PathsParameter:
    return PathsParameter(
        default=default,
        default_factory=default_factory,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
    )


def Body(
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
    alias: Optional[str] = None,
    embed: bool = False,
) -> BodyParameter:
    return BodyParameter(
        alias=alias,
        default=default,
        default_factory=default_factory,
        embed=embed,
    )


def Depends(
    dependency: Optional[Callable] = None,
    /,
    *,
    use_cache: bool = True,
) -> DependencyParameter:
    return DependencyParameter(dependency=dependency, use_cache=use_cache)


def URL() -> URLParameter:
    return URLParameter()


def Request() -> RequestParameter:
    return RequestParameter()


def Response() -> ResponseParameter:
    return ResponseParameter()


def StatusCode() -> StatusCodeParameter:
    return StatusCodeParameter()


def Reason() -> ReasonParameter:
    return ReasonParameter()


def State(
    name: Optional[str] = None,
    *,
    default: Any = Required,
    default_factory: Optional[Supplier[Any]] = None,
) -> StateParameter:
    return StateParameter(
        default=default,
        default_factory=default_factory,
        alias=name,
    )



from typing import Any, Callable, Optional, Sequence, Type, TypeVar

from mediate.protocols import MiddlewareCallable

from ..errors import CompositionError
from ..middlewares import (
    ExpectedContentTypeMiddleware,
    ExpectedHeaderMiddleware,
    ExpectedStatusCodeMiddleware,
)
from ..middlewares import raise_for_status as raise_for_status_middleware
from ..models import Request, Response
from ..operation import Operation, get_operation
from ..services import Service
from ..specification import ClientSpecification

__all__: Sequence[str] = (
    "middleware",
    "expect_content_type",
    "expect_header",
    "expect_status",
    "raise_for_status",
)

C = TypeVar("C", bound=Callable[..., Any])
S = TypeVar("S", bound=Type[Service])
CS = TypeVar("CS", Callable[..., Any], Type[Service])


def middleware(
    *middleware: MiddlewareCallable[Request, Response]
) -> Callable[[CS], CS]:
    def decorate(target: CS, /) -> CS:
        if isinstance(target, type):
            if not issubclass(target, Service):
                raise CompositionError(f"Target class is not a subclass of {Service}")

            client_specification: ClientSpecification = target._spec

            client_specification.middleware.add_all(middleware)
        elif callable(target):
            operation: Operation = get_operation(target)

            operation.middleware.add_all(middleware)
        else:
            raise CompositionError(f"Target of unsupported type {type(target)}")

        return target

    return decorate


def expect_content_type(content_type: str, /) -> Callable[[CS], CS]:
    return middleware(ExpectedContentTypeMiddleware(content_type))


def expect_header(name: str, value: Optional[str] = None) -> Callable[[CS], CS]:
    return middleware(ExpectedHeaderMiddleware(name, value))


def expect_status(*status_codes: int) -> Callable[[CS], CS]:
    return middleware(ExpectedStatusCodeMiddleware(*status_codes))


def raise_for_status(target: CS, /) -> CS:
    return middleware(raise_for_status_middleware)(target)



from abc import abstractmethod
from typing import Any, Callable, Protocol, Sequence, TypeVar, runtime_checkable

from typing_extensions import TypeAlias

from .models import ClientOptions, PreRequest, Request, Response

__all__: Sequence[str] = (
    "AnyCallable",
    "CallNext",
    "Decorator",
    "Dependency",
    "ClientConsumer",
    "Composer",
    "Consumer",
    "Function",
    "RequestConsumer",
    "RequestResolver",
    "ResponseResolver",
    "Supplier",
    "SupportsConsumeClient",
    "SupportsConsumeRequest",
    "SupportsResolveRequest",
    "SupportsResolveResponse",
)

T = TypeVar("T")
T_contra = TypeVar("T_contra", contravariant=True)
T_co = TypeVar("T_co", covariant=True)
R_co = TypeVar("R_co", covariant=True)

AnyCallable: TypeAlias = Callable[..., Any]
Dependency: TypeAlias = AnyCallable


class CallNext(Protocol):
    def __call__(self, request: Request, /) -> Response:
        ...


class Decorator(Protocol[T]):
    @abstractmethod
    def __call__(self, target: T, /) -> T:
        ...


class Supplier(Protocol[T_co]):
    def __call__(self) -> T_co:
        ...


class Consumer(Protocol[T_contra]):
    def __call__(self, t: T_contra, /) -> None:
        ...


class Function(Protocol[T_contra, R_co]):
    def __call__(self, t: T_contra, /) -> R_co:
        ...


class ResponseResolver(Function[Response, T_co], Protocol[T_co]):
    pass


class RequestResolver(Function[PreRequest, T_co], Protocol[T_co]):
    pass


class RequestConsumer(Consumer[PreRequest], Protocol):
    pass


class ClientConsumer(Consumer[ClientOptions], Protocol):
    pass


class Composer(Protocol):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        ...


@runtime_checkable
class SupportsConsumeRequest(Protocol):
    @abstractmethod
    def consume_request(self, request: PreRequest, /) -> None:
        ...


@runtime_checkable
class SupportsConsumeClient(Protocol):
    @abstractmethod
    def consume_client(self, client: ClientOptions, /) -> None:
        ...


@runtime_checkable
class SupportsResolveRequest(Protocol[T_co]):
    @abstractmethod
    def resolve_request(self, request: PreRequest, /) -> T_co:
        ...


@runtime_checkable
class SupportsResolveResponse(Protocol[T_co]):
    @abstractmethod
    def resolve_response(self, response: Response, /) -> T_co:
        ...



from dataclasses import dataclass, field
from typing import MutableSequence, Optional, Sequence

from .middlewares import Middleware
from .models import ClientOptions
from .typing import Dependency

__all__: Sequence[str] = ("ClientSpecification",)


@dataclass
class ClientSpecification:
    options: ClientOptions = field(default_factory=ClientOptions)
    middleware: Middleware = field(default_factory=Middleware)
    default_response: Optional[Dependency] = None
    request_dependencies: MutableSequence[Dependency] = field(default_factory=list)
    response_dependencies: MutableSequence[Dependency] = field(default_factory=list)



from dataclasses import dataclass
from typing import Any, Callable, Sequence, Type, TypeVar, Union

from ..errors import CompositionError
from ..models import ClientOptions, PreRequest
from ..operation import get_operation
from ..services import Service
from ..typing import SupportsConsumeClient, SupportsConsumeRequest

__all__: Sequence[str] = (
    "ConsumerDecorator",
    "OperationConsumerDecorator",
    "ServiceConsumerDecorator",
)

C = TypeVar("C", bound=Callable[..., Any])
S = TypeVar("S", bound=Type[Service])
CS = TypeVar("CS", Callable[..., Any], Type[Service])


@dataclass
class OperationConsumerDecorator:
    consumer: Union[SupportsConsumeRequest, SupportsConsumeClient]

    def __call__(self, target: C, /) -> C:
        if not callable(target):
            raise CompositionError(f"Target of unsupported type {type(target)}")

        if isinstance(self.consumer, SupportsConsumeRequest):
            pre_request: PreRequest = get_operation(target).request_options

            self.consumer.consume_request(pre_request)
        else:
            client_options: ClientOptions = get_operation(target).client_options

            self.consumer.consume_client(client_options)

        return target


@dataclass
class ServiceConsumerDecorator:
    consumer: SupportsConsumeClient

    def __call__(self, target: S, /) -> S:
        if not isinstance(target, type):
            raise CompositionError(f"Target of unsupported type {type(target)}")

        if not issubclass(target, Service):
            raise CompositionError(f"Target class is not a subclass of {Service}")

        client: ClientOptions = target._spec.options

        self.consumer.consume_client(client)

        return target


@dataclass
class ConsumerDecorator:
    consumer: Union[SupportsConsumeRequest, SupportsConsumeClient]

    def __call__(self, target: CS, /) -> CS:
        if isinstance(target, type):
            if not isinstance(self.consumer, SupportsConsumeClient):
                raise CompositionError(
                    f"Consumer {type(self.consumer).__name__!r} does not support consumption"
                    f" of type {Service}"
                )

            return ServiceConsumerDecorator(self.consumer)(target)
        elif callable(target):
            return OperationConsumerDecorator(self.consumer)(target)
        else:
            raise CompositionError(f"Target of unsupported type {type(target)}")



from enum import Enum, auto
from typing import Sequence

__all__: Sequence[str] = (
    "Entity",
    "HTTPMethod",
    "HTTPHeader",
)


class HiddenValueEnum(Enum):
    def __repr__(self) -> str:
        return f"<{type(self).__name__}.{self.name}>"


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value


class Entity(HiddenValueEnum):
    MIDDLEWARE = auto()
    RESPONSE = auto()
    REQUEST_DEPENDENCY = auto()
    RESPONSE_DEPENDENCY = auto()


class HTTPMethod(HiddenValueEnum, StrEnum):
    CONNECT = "CONNECT"
    DELETE = "DELETE"
    GET = "GET"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    PATCH = "PATCH"
    POST = "POST"
    PUT = "PUT"
    TRACE = "TRACE"


class HTTPHeader(HiddenValueEnum, StrEnum):
    ACCEPT = "Accept"
    ACCEPT_ENCODING = "Accept-Encoding"
    ACCEPT_LANGUAGE = "Accept-Language"
    CACHE_CONTROL = "Cache-Control"
    CONNECTION = "Connection"
    CONTENT_TYPE = "Content-Type"
    COOKIE = "Cookie"
    DNT = "DNT"
    HOST = "Host"
    KEEP_ALIVE = "Keep-Alive"
    LOCATION = "Location"
    PRAGMA = "Pragma"
    PROXY_AUTHENTICATE = "Proxy-Authenticate"
    PROXY_AUTHORIZATION = "Proxy-Authorization"
    REFERER = "Referer"
    SERVER = "Server"
    TE = "TE"
    TRAILER = "Trailer"
    TRANSFER_ENCODING = "Transfer-Encoding"
    UPGRADE = "Upgrade"
    USER_AGENT = "User-Agent"



from .client import base_url
from .common import (
    accept,
    cookie,
    cookies,
    header,
    headers,
    query,
    query_params,
    referer,
    request_depends,
    response_depends,
    timeout,
    user_agent,
    verify,
)
from .middleware import (
    expect_content_type,
    expect_header,
    expect_status,
    middleware,
    raise_for_status,
)
from .operation import content, data, files, json, mount, path, path_params
from .request import delete, get, head, options, patch, post, put, request
from .response import response
from .service import service
from .utils import persist_pre_request



from typing import Callable, Sequence, Type, TypeVar

from ..consumers import BaseURLConsumer
from ..services import Service
from .api import ConsumerDecorator

__all__: Sequence[str] = ("base_url",)

TT = TypeVar("TT", Callable, Type[Service])


def base_url(base_url: str, /) -> Callable[[TT], TT]:
    return ConsumerDecorator(BaseURLConsumer(base_url))



import collections.abc
import dataclasses
import typing
import urllib.parse
from collections import Counter
from typing import (
    Any,
    Callable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
    Set,
    Tuple,
)

from pydantic import BaseModel
from pydantic.fields import FieldInfo, ModelField

from . import api, utils
from .errors import DuplicateParameters
from .models import PreRequest
from .params import BodyParameter, Parameter, PathParameter, QueryParameter
from .validation import ValidatedFunction

__all__: Sequence[str] = (
    "get_fields",
    "validate_fields",
    "compose",
)


def get_fields(
    request: PreRequest,
    func: Callable,
) -> Mapping[str, Tuple[Any, Parameter]]:
    path_params: Set[str] = (
        utils.parse_format_string(urllib.parse.unquote(str(request.url)))
        if request is not None
        else set()
    )

    fields: MutableMapping[str, Tuple[Any, Parameter]] = {}

    field_name: str
    model_field: ModelField
    for field_name, model_field in ValidatedFunction(func).model.__fields__.items():
        field_info: FieldInfo = model_field.field_info
        parameter: Parameter

        # Parameter Inference
        if not isinstance(field_info, Parameter):
            if field_name in path_params:
                parameter = PathParameter(
                    alias=field_name,
                    default=utils.get_default(field_info),
                )
            elif (
                (
                    isinstance(model_field.annotation, type)
                    and issubclass(model_field.annotation, (BaseModel, dict))
                )
                or dataclasses.is_dataclass(model_field.annotation)
                or (
                    utils.is_generic_alias(model_field.annotation)
                    and typing.get_origin(model_field.annotation)
                    in (collections.abc.Mapping,)
                )
            ):
                parameter = BodyParameter(
                    alias=field_name,
                    default=utils.get_default(field_info),
                )
            else:
                parameter = QueryParameter(
                    default=utils.get_default(field_info),
                )
        else:
            parameter = field_info

        # Create a clone of the parameter so that any mutations do not affect the original
        parameter_clone: Parameter = dataclasses.replace(parameter)

        parameter_clone.prepare(model_field)

        fields[field_name] = (model_field.annotation, parameter_clone)

    total_body_fields: int = sum(
        isinstance(parameter, BodyParameter) for _, parameter in fields.values()
    )

    if total_body_fields > 1:
        field: str
        annotation: Any
        param: Parameter
        for field, (annotation, param) in fields.items():
            if not isinstance(param, BodyParameter):
                continue

            param = dataclasses.replace(param, embed=True)

            fields[field] = (annotation, param)

    return fields


def validate_fields(fields: Mapping[str, Tuple[Any, Parameter]], /) -> None:
    parameter_aliases: MutableSequence[str] = [
        parameter.alias
        for _, parameter in fields.values()
        if parameter.alias is not None
    ]

    # Validate that there are no parameters using the same alias
    #   For example, the following function should fail validation:
    #       @get("/")
    #       def foo(a: str = Query(alias="name"), b: str = Query(alias="name")): ...
    alias_counts: Mapping[str, int] = Counter(parameter_aliases)
    duplicate_aliases: Set[str] = {
        alias for alias, count in alias_counts.items() if count > 1
    }
    if duplicate_aliases:
        raise DuplicateParameters(f"Duplicate parameters: {duplicate_aliases!r}")


def compose(
    func: Callable,
    request: PreRequest,
    args: Tuple[Any, ...],
    kwargs: Mapping[str, Any],
) -> None:
    arguments: Mapping[str, Any] = api.bind_arguments(func, args, kwargs)

    fields: Mapping[str, Tuple[Any, Parameter]] = get_fields(request, func)

    # Validate that the fields are acceptable
    validate_fields(fields)

    model: BaseModel = api.create_model(func, fields, arguments)

    # By this stage the arguments have been validated
    validated_arguments: Mapping[str, Any] = model.dict()

    field_name: str
    parameter: Parameter
    for field_name, (_, parameter) in fields.items():
        argument: Any = validated_arguments[field_name]

        parameter.compose(request, argument)

    # Validate the request (e.g. to ensure no path params have been missed)
    request.validate()



from typing import Any, Callable, Optional, TypeVar, overload

from .typing import Supplier

T = TypeVar("T")

@overload
def Query(
    name: Optional[str] = None,
    *,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> Any: ...
@overload
def Query(
    name: Optional[str] = None,
    *,
    default: T,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> Any: ...
@overload
def Query(
    name: Optional[str] = None,
    *,
    default_factory: Supplier[T],
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> T: ...
@overload
def Header(
    name: Optional[str] = None,
    *,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    convert_underscores: bool = True,
) -> Any: ...
@overload
def Header(
    name: Optional[str] = None,
    *,
    default: T,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    convert_underscores: bool = True,
) -> T: ...
@overload
def Header(
    name: Optional[str] = None,
    *,
    default_factory: Supplier[T],
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    convert_underscores: bool = True,
) -> T: ...
@overload
def Cookie(
    name: Optional[str] = None,
    *,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> Any: ...
@overload
def Cookie(
    name: Optional[str] = None,
    *,
    default: T,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> T: ...
@overload
def Cookie(
    name: Optional[str] = None,
    *,
    default_factory: Supplier[T],
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> T: ...
@overload
def Path(
    name: Optional[str] = None,
    *,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> Any: ...
@overload
def Path(
    name: Optional[str] = None,
    *,
    default: T,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> T: ...
@overload
def Path(
    name: Optional[str] = None,
    *,
    default_factory: Supplier[T],
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
) -> T: ...

# Queries
@overload
def Queries(
    *,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> Any: ...
@overload
def Queries(
    default: T,
    *,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> T: ...
@overload
def Queries(
    *,
    default_factory: Supplier[T],
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> T: ...

# Headers
@overload
def Headers(
    *,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> Any: ...
@overload
def Headers(
    default: T,
    *,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> T: ...
@overload
def Headers(
    *,
    default_factory: Supplier[T],
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> T: ...

# Cookies
@overload
def Cookies(
    *,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> Any: ...
@overload
def Cookies(
    default: T,
    *,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> T: ...
@overload
def Cookies(
    *,
    default_factory: Supplier[T],
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> T: ...

# Paths
@overload
def Paths(
    *,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> Any: ...
@overload
def Paths(
    default: T,
    *,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> T: ...
@overload
def Paths(
    *,
    default_factory: Supplier[T],
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
) -> T: ...

# Body
@overload
def Body(
    *,
    alias: Optional[str] = None,
    embed: bool = False,
) -> Any: ...
@overload
def Body(
    default: T,
    *,
    alias: Optional[str] = None,
    embed: bool = False,
) -> T: ...
@overload
def Body(
    *,
    default_factory: Callable[[], T],
    alias: Optional[str] = None,
    embed: bool = False,
) -> T: ...

# Depends
@overload
def Depends(*, use_cache: bool = True) -> Any: ...
@overload
def Depends(dependency: Callable[..., T], /, *, use_cache: bool = True) -> T: ...

# URL
def URL() -> Any: ...

# Request
def Request() -> Any: ...

# Response
def Response() -> Any: ...

# StatusCode
def StatusCode() -> Any: ...

# Reason
def Reason() -> Any: ...

# State
@overload
def State(name: Optional[str] = None) -> Any: ...
@overload
def State(name: Optional[str] = None, *, default: T) -> Any: ...
@overload
def State(name: Optional[str] = None, *, default_factory: Supplier[T]) -> T: ...



from typing import Any, Callable

from .dependence import DependencyParameter
from .models import PreRequest, Response


def resolve_response(func: Callable, response: Response) -> Any:
    dependency: DependencyParameter = DependencyParameter(dependency=func)

    return dependency.resolve_response(response)


def resolve_request(func: Callable, request: PreRequest) -> Any:
    dependency: DependencyParameter = DependencyParameter(dependency=func)

    return dependency.resolve_request(request)



from typing import Any, Callable, Mapping, Sequence, Tuple, Type

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from . import utils
from .validation import create_func_model

__all__: Sequence[str] = (
    "create_model_cls",
    "create_model",
    "bind_arguments",
)


def create_model_cls(
    func: Callable, fields: Mapping[str, Tuple[Any, FieldInfo]]
) -> Type[BaseModel]:
    class Config:
        allow_population_by_field_name: bool = True
        arbitrary_types_allowed: bool = True

    return create_func_model(func, fields, config=Config)


def create_model(
    func: Callable,
    fields: Mapping[str, Tuple[Any, FieldInfo]],
    arguments: Mapping[str, Any],
) -> BaseModel:
    model_cls: Type[BaseModel] = create_model_cls(func, fields)

    return model_cls(**arguments)


def bind_arguments(
    func: Callable, args: Tuple[Any, ...], kwargs: Mapping[str, Any]
) -> Mapping[str, Any]:
    arguments: Mapping[str, Any] = utils.bind_arguments(func, args, kwargs)

    return {
        key: value
        for key, value in arguments.items()
        if not isinstance(value, FieldInfo)
    }



from dataclasses import dataclass
from typing import Any, Optional, Sequence, TypeVar

from httpx import Cookies, Headers, QueryParams

from .models import PreRequest, Response, State
from .typing import ResponseResolver, SupportsResolveRequest, SupportsResolveResponse

__all__: Sequence[str] = (
    "BodyResolver",
    "CookieResolver",
    "CookiesResolver",
    "HeaderResolver",
    "HeadersResolver",
    "QueryResolver",
    "QueriesResolver",
    "StateResolver",
)

T = TypeVar("T")


@dataclass
class QueryResolver(SupportsResolveRequest, SupportsResolveResponse):
    name: str

    def resolve_request(self, request: PreRequest, /) -> Optional[Sequence[str]]:
        return self.resolve(request.params)

    def resolve_response(self, response: Response, /) -> Optional[Sequence[str]]:
        return self.resolve(response.request.url.params)

    def resolve(self, query_params: QueryParams, /) -> Optional[Sequence[str]]:
        if self.name in query_params:
            return query_params.get_list(self.name)

        return None


@dataclass
class HeaderResolver(SupportsResolveRequest, SupportsResolveResponse):
    name: str

    def resolve_request(self, request: PreRequest, /) -> Optional[Sequence[str]]:
        return self.resolve(request.headers)

    def resolve_response(self, response: Response, /) -> Optional[Sequence[str]]:
        return self.resolve(response.headers)

    def resolve(self, headers: Headers, /) -> Optional[Sequence[str]]:
        if self.name in headers:
            return headers.get_list(self.name)

        return None


@dataclass
class CookieResolver(SupportsResolveRequest, SupportsResolveResponse):
    name: str

    def resolve_request(self, request: PreRequest, /) -> Optional[str]:
        return self.resolve(request.cookies)

    def resolve_response(self, response: Response, /) -> Optional[str]:
        return self.resolve(response.cookies)

    def resolve(self, cookies: Cookies, /) -> Optional[str]:
        return cookies.get(self.name)


class QueriesResolver(SupportsResolveRequest, SupportsResolveResponse):
    @staticmethod
    def resolve_request(request: PreRequest, /) -> QueryParams:
        return request.params

    @staticmethod
    def resolve_response(response: Response, /) -> QueryParams:
        return response.request.url.params


class HeadersResolver(SupportsResolveRequest, SupportsResolveResponse):
    @staticmethod
    def resolve_request(request: PreRequest, /) -> Headers:
        return request.headers

    @staticmethod
    def resolve_response(response: Response, /) -> Headers:
        return response.headers


class CookiesResolver(SupportsResolveRequest, SupportsResolveResponse):
    @staticmethod
    def resolve_request(request: PreRequest, /) -> Cookies:
        return request.cookies

    @staticmethod
    def resolve_response(response: Response, /) -> Cookies:
        return response.cookies


class BodyResolver(ResponseResolver[Any]):
    @staticmethod
    def __call__(response: Response, /) -> Any:
        return response.json()


@dataclass
class StateResolver(SupportsResolveRequest, SupportsResolveResponse):
    key: str

    def resolve_request(self, request: PreRequest, /) -> Any:
        return self.resolve(request.state)

    def resolve_response(self, response: Response, /) -> Any:
        return self.resolve(response.state)

    def resolve(self, state: State, /) -> Any:
        return state.get(self.key)



from dataclasses import dataclass
from typing import Any, Mapping, Sequence, Tuple

from httpx import URL, Cookies, Headers, QueryParams, Timeout

from . import converters
from .models import ClientOptions, PreRequest
from .types import (
    CookiesTypes,
    CookieTypes,
    HeadersTypes,
    HeaderTypes,
    JsonTypes,
    PathsTypes,
    PathTypes,
    QueriesTypes,
    QueryTypes,
    RequestContent,
    RequestData,
    RequestFiles,
    TimeoutTypes,
    VerifyTypes,
)
from .typing import SupportsConsumeClient, SupportsConsumeRequest

__all__: Sequence[str] = (
    "QueryConsumer",
    "HeaderConsumer",
    "CookieConsumer",
    "PathConsumer",
    "QueriesConsumer",
    "HeadersConsumer",
    "CookiesConsumer",
    "PathsConsumer",
    "ContentConsumer",
    "DataConsumer",
    "FilesConsumer",
    "JsonConsumer",
    "TimeoutConsumer",
    "MountConsumer",
    "BaseURLConsumer",
    "VerifyConsumer",
)


@dataclass(init=False)
class QueryConsumer(SupportsConsumeClient, SupportsConsumeRequest):
    key: str
    values: Sequence[str]

    def __init__(self, key: str, value: QueryTypes) -> None:
        self.key = key
        self.values = converters.convert_query_param(value)

    def consume_request(self, request: PreRequest, /) -> None:
        request.params = self._apply(request.params)

    def consume_client(self, client: ClientOptions, /) -> None:
        client.params = self._apply(client.params)

    def _apply(self, params: QueryParams, /) -> QueryParams:
        # If there's only one value, set the query param and overwrite any
        # existing entries for this key
        if len(self.values) == 1:
            return params.set(self.key, self.values[0])

        # Otherwise, update the query params and maintain any existing entries for
        # this key
        value: str
        for value in self.values:
            params = params.add(self.key, value)

        return params


@dataclass(init=False)
class HeaderConsumer(SupportsConsumeRequest, SupportsConsumeClient):
    key: str
    values: Sequence[str]

    def __init__(self, key: str, value: HeaderTypes) -> None:
        self.key = key
        self.values = converters.convert_header(value)

    def consume_request(self, request: PreRequest, /) -> None:
        self._apply(request.headers)

    def consume_client(self, client: ClientOptions, /) -> None:
        self._apply(client.headers)

    def _apply(self, headers: Headers, /) -> None:
        # If there's only one value, set the header and overwrite any existing
        # entries for this key
        if len(self.values) == 1:
            headers[self.key] = self.values[0]
        # Otherwise, update the headers and maintain any existing entries for this
        # key
        else:
            values: Sequence[Tuple[str, str]] = [
                (self.key, value) for value in self.values
            ]

            headers.update(values)


@dataclass(init=False)
class CookieConsumer(SupportsConsumeRequest, SupportsConsumeClient):
    key: str
    value: str

    def __init__(self, key: str, value: CookieTypes) -> None:
        self.key = key
        self.value = converters.convert_cookie(value)

    def consume_request(self, request: PreRequest, /) -> None:
        request.cookies[self.key] = self.value

    def consume_client(self, client: ClientOptions, /) -> None:
        client.cookies[self.key] = self.value


@dataclass(init=False)
class PathConsumer(SupportsConsumeRequest):
    key: str
    value: str

    def __init__(self, key: str, value: PathTypes) -> None:
        self.key = key
        self.value = converters.convert_path_param(value)

    def consume_request(self, request: PreRequest, /) -> None:
        request.path_params[self.key] = self.value


@dataclass(init=False)
class QueriesConsumer(SupportsConsumeRequest, SupportsConsumeClient):
    params: QueryParams

    def __init__(self, params: QueriesTypes, /) -> None:
        self.params = converters.convert_query_params(params)

    def consume_request(self, request: PreRequest, /) -> None:
        request.params = request.params.merge(self.params)

    def consume_client(self, client: ClientOptions, /) -> None:
        client.params = client.params.merge(self.params)


@dataclass(init=False)
class HeadersConsumer(SupportsConsumeRequest, SupportsConsumeClient):
    headers: Headers

    def __init__(self, headers: HeadersTypes, /) -> None:
        self.headers = converters.convert_headers(headers)

    def consume_request(self, request: PreRequest, /) -> None:
        request.headers.update(self.headers)

    def consume_client(self, client: ClientOptions, /) -> None:
        client.headers.update(self.headers)


@dataclass(init=False)
class CookiesConsumer(SupportsConsumeRequest, SupportsConsumeClient):
    cookies: Cookies

    def __init__(self, cookies: CookiesTypes, /) -> None:
        self.cookies = converters.convert_cookies(cookies)

    def consume_request(self, request: PreRequest, /) -> None:
        request.cookies.update(self.cookies)

    def consume_client(self, client: ClientOptions, /) -> None:
        client.cookies.update(self.cookies)


@dataclass(init=False)
class PathsConsumer(SupportsConsumeRequest):
    path_params: Mapping[str, str]

    def __init__(self, path_params: PathsTypes, /) -> None:
        self.path_params = converters.convert_path_params(path_params)

    def consume_request(self, request: PreRequest, /) -> None:
        request.path_params.update(self.path_params)


@dataclass
class ContentConsumer(SupportsConsumeRequest):
    content: RequestContent

    def consume_request(self, request: PreRequest, /) -> None:
        request.content = self.content


@dataclass
class DataConsumer(SupportsConsumeRequest):
    data: RequestData

    def consume_request(self, request: PreRequest, /) -> None:
        request.data = self.data


@dataclass
class FilesConsumer(SupportsConsumeRequest):
    files: RequestFiles

    def consume_request(self, request: PreRequest, /) -> None:
        request.files = self.files


@dataclass
class JsonConsumer(SupportsConsumeRequest):
    json: JsonTypes

    def consume_request(self, request: PreRequest, /) -> None:
        request.json = self.json


@dataclass(init=False)
class TimeoutConsumer(SupportsConsumeRequest, SupportsConsumeClient):
    timeout: Timeout

    def __init__(self, timeout: TimeoutTypes, /) -> None:
        self.timeout = converters.convert_timeout(timeout)

    def consume_request(self, request: PreRequest, /) -> None:
        request.timeout = self.timeout

    def consume_client(self, client: ClientOptions, /) -> None:
        client.timeout = self.timeout


@dataclass
class StateConsumer(SupportsConsumeRequest):
    key: str
    value: Any

    def consume_request(self, request: PreRequest, /) -> None:
        request.state[self.key] = self.value


@dataclass
class MountConsumer(SupportsConsumeRequest):
    path: str

    def consume_request(self, request: PreRequest, /) -> None:
        request.url = request.url.copy_with(path=self.path + request.url.path)


@dataclass
class BaseURLConsumer(SupportsConsumeClient):
    base_url: str

    def consume_client(self, client: ClientOptions, /) -> None:
        client.base_url = URL(self.base_url)


@dataclass
class VerifyConsumer(SupportsConsumeClient):
    verify: VerifyTypes

    def consume_client(self, client: ClientOptions, /) -> None:
        client.verify = self.verify



import functools
import inspect
from dataclasses import dataclass, field
from json import JSONDecodeError
from typing import Any, Callable, Generic, MutableSequence, Optional, Sequence, TypeVar

import httpx
import pydantic
from httpx import Client
from pydantic import BaseModel
from typing_extensions import ParamSpec

from .composition import compose
from .errors import NotAnOperationError
from .middlewares import Middleware
from .models import ClientOptions, PreRequest, Request, Response
from .resolution import resolve_request, resolve_response
from .typing import Dependency

__all__: Sequence[str] = (
    "set_operation",
    "has_operation",
    "get_operation",
    "Operation",
)

PS = ParamSpec("PS")
RT_co = TypeVar("RT_co", covariant=True)

ATTRIBUTE_OPERATION: str = "operation"


def set_operation(func: Callable, operation: "Operation", /) -> None:
    setattr(func, ATTRIBUTE_OPERATION, operation)


def has_operation(func: Callable, /) -> bool:
    return hasattr(func, ATTRIBUTE_OPERATION)


def get_operation(func: Callable, /) -> "Operation":
    operation: Optional[Operation] = getattr(func, ATTRIBUTE_OPERATION, None)

    if operation is None:
        raise NotAnOperationError(f"{func!r} is not an operation")

    return operation


@dataclass
class Operation(Generic[PS, RT_co]):
    func: Callable[PS, RT_co]
    client_options: ClientOptions
    request_options: PreRequest
    client: Optional[Client] = None
    response: Optional[Dependency] = None
    middleware: Middleware = field(default_factory=Middleware)
    request_dependencies: MutableSequence[Dependency] = field(default_factory=list)
    response_dependencies: MutableSequence[Dependency] = field(default_factory=list)

    def __call__(self, *args: PS.args, **kwargs: PS.kwargs) -> Any:
        client: Client

        if self.client is not None:
            client = self.client
        else:
            # Build a disposable client using the available client options
            client = self.client_options.build()

        # Create a clone of the request options, so that mutations don't
        # affect the original copy.
        # Mutations to the request options will occur during composition.
        pre_request: PreRequest = self.request_options.clone()

        # Compose the request using the provided arguments
        compose(self.func, pre_request, args, kwargs)

        # Compose the request using each of the composition dependencies
        request_dependency: Dependency
        for request_dependency in self.request_dependencies:
            resolve_request(request_dependency, pre_request)

        request: Request = pre_request.build_request(client)

        return_annotation: Any = inspect.signature(self.func).return_annotation

        if return_annotation is PreRequest:
            return pre_request
        if return_annotation is Request:
            return request

        @self.middleware.compose
        def send_request(request: Request, /) -> Response:
            httpx_response: httpx.Response = client.send(request)

            return Response.from_httpx_response(httpx_response)

        response: Response = send_request(request)

        # Feed the response through each of the response dependencies
        response_dependency: Dependency
        for response_dependency in self.response_dependencies:
            resolve_response(response_dependency, response)

        if self.response is not None:
            return resolve_response(self.response, response)

        if return_annotation is inspect.Parameter.empty:
            try:
                return response.json()
            except JSONDecodeError:
                return response.text
        if return_annotation is None:
            return None
        if return_annotation is Response:
            return response
        if isinstance(return_annotation, type) and issubclass(
            return_annotation, BaseModel
        ):
            return return_annotation.parse_obj(response.json())

        return pydantic.parse_raw_as(return_annotation, response.text)

    @property
    def wrapper(self) -> Callable[PS, RT_co]:
        @functools.wraps(self.func)
        def wrapper(*args: PS.args, **kwargs: PS.kwargs) -> RT_co:
            if inspect.ismethod(self.func):
                # Read off `self` or `cls`
                _, *args = args  # type: ignore

            return self(*args, **kwargs)

        set_operation(wrapper, self)

        return wrapper



from typing import Callable, Sequence, Type, TypeVar

from ..consumers import (
    CookieConsumer,
    CookiesConsumer,
    HeaderConsumer,
    HeadersConsumer,
    QueriesConsumer,
    QueryConsumer,
    TimeoutConsumer,
    VerifyConsumer,
)
from ..enums import HTTPHeader
from ..errors import CompositionError
from ..operation import Operation, get_operation
from ..services import ClientSpecification, Service
from ..types import (
    CookiesTypes,
    CookieTypes,
    HeadersTypes,
    HeaderTypes,
    QueriesTypes,
    QueryTypes,
    TimeoutTypes,
    VerifyTypes,
)
from ..typing import Dependency
from .api import ConsumerDecorator

__all__: Sequence[str] = (
    "accept",
    "cookie",
    "cookies",
    "request_depends",
    "response_depends",
    "header",
    "headers",
    "query",
    "query_params",
    "referer",
    "timeout",
    "user_agent",
    "verify",
)

TT = TypeVar("TT", Callable, Type[Service])


def accept(*content_types: str) -> Callable[[TT], TT]:
    return ConsumerDecorator(
        HeaderConsumer(
            HTTPHeader.ACCEPT,
            ",".join(content_types),
        )
    )


def request_depends(*dependencies: Dependency) -> Callable[[TT], TT]:
    def decorate(target: TT, /) -> TT:
        if isinstance(target, type):
            if not issubclass(target, Service):
                raise CompositionError(f"Target class is not a subclass of {Service}")

            client_specification: ClientSpecification = target._spec

            client_specification.request_dependencies.extend(dependencies)
        elif callable(target):
            operation: Operation = get_operation(target)

            operation.request_dependencies.extend(dependencies)
        else:
            raise CompositionError(f"Target of unsupported type {type(target)}")

        return target

    return decorate


def response_depends(*dependencies: Dependency) -> Callable[[TT], TT]:
    def decorate(target: TT, /) -> TT:
        if isinstance(target, type):
            if not issubclass(target, Service):
                raise CompositionError(f"Target class is not a subclass of {Service}")

            client_specification: ClientSpecification = target._spec

            client_specification.response_dependencies.extend(dependencies)
        elif callable(target):
            operation: Operation = get_operation(target)

            operation.response_dependencies.extend(dependencies)
        else:
            raise CompositionError(f"Target of unsupported type {type(target)}")

        return target

    return decorate


def cookie(key: str, value: CookieTypes) -> Callable[[TT], TT]:
    return ConsumerDecorator(CookieConsumer(key, value))


def cookies(cookies: CookiesTypes, /) -> Callable[[TT], TT]:
    return ConsumerDecorator(CookiesConsumer(cookies))


def header(key: str, value: HeaderTypes) -> Callable[[TT], TT]:
    return ConsumerDecorator(HeaderConsumer(key, value))


def headers(headers: HeadersTypes, /) -> Callable[[TT], TT]:
    return ConsumerDecorator(HeadersConsumer(headers))


def query(key: str, value: QueryTypes) -> Callable[[TT], TT]:
    return ConsumerDecorator(QueryConsumer(key, value))


def query_params(params: QueriesTypes, /) -> Callable[[TT], TT]:
    return ConsumerDecorator(QueriesConsumer(params))


def referer(referer: str, /) -> Callable[[TT], TT]:
    return ConsumerDecorator(
        HeaderConsumer(
            HTTPHeader.REFERER,
            referer,
        )
    )


def timeout(timeout: TimeoutTypes, /) -> Callable[[TT], TT]:
    return ConsumerDecorator(TimeoutConsumer(timeout))


def user_agent(user_agent: str, /) -> Callable[[TT], TT]:
    return ConsumerDecorator(
        HeaderConsumer(
            HTTPHeader.USER_AGENT,
            user_agent,
        )
    )


def verify(verify: VerifyTypes, /) -> Callable[[TT], TT]:
    return ConsumerDecorator(VerifyConsumer(verify))



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, Generic, Mapping, Optional, Sequence, Set, TypeVar, Union

import fastapi.encoders
import httpx
from httpx import Cookies, Headers, QueryParams
from pydantic import Required
from pydantic.fields import FieldInfo, ModelField

from .consumers import (
    CookieConsumer,
    CookiesConsumer,
    HeaderConsumer,
    HeadersConsumer,
    PathConsumer,
    PathsConsumer,
    QueriesConsumer,
    QueryConsumer,
    StateConsumer,
)
from .converters import (
    convert_cookie,
    convert_header,
    convert_path_param,
    convert_query_param,
)
from .errors import CompositionError, ResolutionError
from .models import PreRequest, Response
from .resolvers import (
    BodyResolver,
    CookieResolver,
    CookiesResolver,
    HeaderResolver,
    HeadersResolver,
    QueriesResolver,
    QueryResolver,
    StateResolver,
)
from .types import CookiesTypes, HeadersTypes, PathsTypes, QueriesTypes
from .typing import RequestConsumer, RequestResolver, ResponseResolver, Supplier
from .utils import parse_obj_as

__all__: Sequence[str] = (
    "QueryParameter",
    "HeaderParameter",
    "CookieParameter",
    "PathParameter",
    "QueriesParameter",
    "HeadersParameter",
    "CookiesParameter",
    "PathsParameter",
    "BodyParameter",
    "URLParameter",
    "ResponseParameter",
    "RequestParameter",
    "StatusCodeParameter",
    "ReasonParameter",
)

K = TypeVar("K")
V = TypeVar("V")


@dataclass(unsafe_hash=True)
class Parameter(FieldInfo):
    alias: Optional[str] = None
    default: Any = Required
    default_factory: Optional[Supplier[Any]] = None
    title: Optional[str] = field(default=None, compare=False)
    description: Optional[str] = field(default=None, compare=False)
    exclude: Union[Set[Union[int, str]], Mapping[Union[int, str], Any], Any] = field(
        default=None, compare=False
    )
    include: Union[Set[Union[int, str]], Mapping[Union[int, str], Any], Any] = field(
        default=None, compare=False
    )
    const: Optional[bool] = field(default=None, compare=False)
    gt: Optional[float] = field(default=None, compare=False)
    ge: Optional[float] = field(default=None, compare=False)
    lt: Optional[float] = field(default=None, compare=False)
    le: Optional[float] = field(default=None, compare=False)
    multiple_of: Optional[float] = field(default=None, compare=False)
    allow_inf_nan: Optional[bool] = field(default=None, compare=False)
    max_digits: Optional[int] = field(default=None, compare=False)
    decimal_places: Optional[int] = field(default=None, compare=False)
    min_items: Optional[int] = field(default=None, compare=False)
    max_items: Optional[int] = field(default=None, compare=False)
    unique_items: Optional[bool] = field(default=None, compare=False)
    min_length: Optional[int] = field(default=None, compare=False)
    max_length: Optional[int] = field(default=None, compare=False)
    allow_mutation: bool = field(default=True, compare=False)
    regex: Optional[str] = field(default=None, compare=False)
    discriminator: Optional[str] = field(default=None, compare=False)
    repr: bool = field(default=True, compare=False)
    extra: Dict[str, Any] = field(default_factory=dict, compare=False)
    alias_priority: Optional[int] = field(
        init=False, repr=False, default=None, compare=False
    )

    def __post_init__(self) -> None:
        # Pydantic validates that the alias is a "strict" string. This means
        # that enums (e.g. `HeaderName`) are unable to be used.
        # To mitigate this, the alias is explicity converted to a string here.
        if self.alias is not None:
            self.alias = str(self.alias)

    def compose(self, request: PreRequest, argument: Any, /) -> None:
        raise CompositionError(f"Parameter {type(self)!r} is not composable")

    def resolve_response(self, response: Response, /) -> Any:
        raise ResolutionError(
            f"Parameter {type(self)!r} is not resolvable for type {Response!r}"
        )

    def resolve_request(self, request: PreRequest, /) -> Any:
        raise ResolutionError(
            f"Parameter {type(self)!r} is not resolvable for type {PreRequest!r}"
        )

    def prepare(self, model_field: ModelField, /) -> None:
        if self.alias is None:
            self.alias = model_field.name


class ComposableSingletonParameter(ABC, Parameter, Generic[K, V]):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        if self.alias is None:
            raise CompositionError(
                f"Cannot compose parameter {type(self)!r} without an alias"
            )

        if argument is None and self.default is not Required:
            return

        consumer: RequestConsumer = self.build_consumer(
            self.parse_key(self.alias),
            self.parse_value(argument),
        )

        consumer(request)

    @abstractmethod
    def parse_key(self, key: str, /) -> K:
        ...

    @abstractmethod
    def parse_value(self, value: Any, /) -> V:
        ...

    @abstractmethod
    def build_consumer(self, key: K, value: V) -> RequestConsumer:
        ...


class ResolvableSingletonParameter(ABC, Parameter, Generic[K, V]):
    def resolve_request(self, request: PreRequest, /) -> V:
        if self.alias is None:
            raise ResolutionError(
                f"Cannot resolve parameter {type(self)!r} without an alias"
            )

        resolver: RequestResolver[V] = self.build_request_resolver(
            self.parse_key(self.alias),
        )

        return resolver(request)

    def resolve_response(self, response: Response, /) -> V:
        if self.alias is None:
            raise ResolutionError(
                f"Cannot resolve parameter {type(self)!r} without an alias"
            )

        resolver: ResponseResolver[V] = self.build_response_resolver(
            self.parse_key(self.alias),
        )

        return resolver(response)

    @abstractmethod
    def parse_key(self, key: str, /) -> K:
        ...

    @abstractmethod
    def build_request_resolver(self, key: K) -> RequestResolver[V]:
        ...

    @abstractmethod
    def build_response_resolver(self, key: K) -> ResponseResolver[V]:
        ...


class ComposableSingletonStringParameter(ComposableSingletonParameter[str, str]):
    def parse_key(self, key: str, /) -> str:
        return key


class ResolvableSingletonStringParameter(
    ResolvableSingletonParameter[str, Optional[str]]
):
    def parse_key(self, key: str, /) -> str:
        return key


class QueryParameter(
    ComposableSingletonParameter[str, Sequence[str]],
    ResolvableSingletonParameter[str, Optional[Sequence[str]]],
):
    def parse_key(self, key: str, /) -> str:
        return key

    def parse_value(self, value: Any, /) -> Sequence[str]:
        return convert_query_param(value)

    def build_consumer(self, key: str, value: Sequence[str]) -> RequestConsumer:
        return QueryConsumer(key, value).consume_request

    def build_request_resolver(
        self, key: str
    ) -> RequestResolver[Optional[Sequence[str]]]:
        return QueryResolver(key).resolve_request

    def build_response_resolver(
        self, key: str
    ) -> ResponseResolver[Optional[Sequence[str]]]:
        return QueryResolver(key).resolve_response


@dataclass(unsafe_hash=True)
class HeaderParameter(
    ComposableSingletonParameter[str, Sequence[str]],
    ResolvableSingletonParameter[str, Optional[Sequence[str]]],
):
    convert_underscores: bool = True

    def parse_key(self, key: str, /) -> str:
        if self.convert_underscores:
            return key.replace("_", "-")

        return key

    def parse_value(self, value: Any, /) -> Sequence[str]:
        return convert_header(value)

    def build_consumer(self, key: str, value: Sequence[str]) -> RequestConsumer:
        return HeaderConsumer(key, value).consume_request

    def build_request_resolver(
        self, key: str
    ) -> RequestResolver[Optional[Sequence[str]]]:
        return HeaderResolver(key).resolve_request

    def build_response_resolver(
        self, key: str
    ) -> ResponseResolver[Optional[Sequence[str]]]:
        return HeaderResolver(key).resolve_response


class CookieParameter(
    ComposableSingletonStringParameter, ResolvableSingletonStringParameter
):
    def parse_value(self, value: Any, /) -> str:
        return convert_cookie(value)

    def build_consumer(self, key: str, value: str) -> RequestConsumer:
        return CookieConsumer(key, value).consume_request

    def build_request_resolver(self, key: str) -> RequestResolver[Optional[str]]:
        return CookieResolver(key).resolve_request

    def build_response_resolver(self, key: str) -> ResponseResolver[Optional[str]]:
        return CookieResolver(key).resolve_response


class PathParameter(ComposableSingletonStringParameter):
    def parse_value(self, value: Any, /) -> str:
        return convert_path_param(value)

    def build_consumer(self, key: str, value: str) -> RequestConsumer:
        return PathConsumer(key, value).consume_request


class QueriesParameter(Parameter):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        params: QueriesTypes = parse_obj_as(QueriesTypes, argument)  # type: ignore

        QueriesConsumer(params).consume_request(request)

    def resolve_request(self, request: PreRequest, /) -> QueryParams:
        return QueriesResolver().resolve_request(request)

    def resolve_response(self, response: Response, /) -> QueryParams:
        return QueriesResolver().resolve_response(response)


class HeadersParameter(Parameter):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        headers: HeadersTypes = parse_obj_as(HeadersTypes, argument)  # type: ignore

        HeadersConsumer(headers).consume_request(request)

    def resolve_request(self, request: PreRequest, /) -> Headers:
        return HeadersResolver().resolve_request(request)

    def resolve_response(self, response: Response, /) -> Headers:
        return HeadersResolver().resolve_response(response)


class CookiesParameter(Parameter):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        cookies: CookiesTypes = parse_obj_as(CookiesTypes, argument)  # type: ignore

        CookiesConsumer(cookies).consume_request(request)

    def resolve_request(self, request: PreRequest, /) -> Cookies:
        return CookiesResolver().resolve_request(request)

    def resolve_response(self, response: Response, /) -> Cookies:
        return CookiesResolver().resolve_response(response)


class PathsParameter(Parameter):
    def compose(self, request: PreRequest, argument: Any, /) -> None:
        path_params: PathsTypes = parse_obj_as(PathsTypes, argument)  # type: ignore

        PathsConsumer(path_params).consume_request(request)


@dataclass(unsafe_hash=True)
class BodyParameter(Parameter):
    embed: bool = False

    def compose(self, request: PreRequest, argument: Any, /) -> None:
        # If the parameter is not required and has no value, it can be omitted
        if argument is None and self.default is not Required:
            return

        json_value: Any = fastapi.encoders.jsonable_encoder(argument)

        if self.embed:
            if self.alias is None:
                raise CompositionError(
                    f"Cannot embed parameter {type(self)!r} without an alias"
                )

            json_value = {self.alias: json_value}

        # If this parameter shouln't be embedded in any pre-existing json,
        # make it the entire JSON request body
        if not self.embed:
            request.json = json_value
        else:
            if request.json is None:
                request.json = json_value
            else:
                request.json.update(json_value)

    def resolve_response(self, response: Response, /) -> Any:
        return BodyResolver()(response)


class URLParameter(Parameter):
    def resolve_request(self, request: PreRequest, /) -> httpx.URL:
        return request.url

    def resolve_response(self, response: Response, /) -> httpx.URL:
        return response.request.url


class ResponseParameter(Parameter):
    def resolve_response(self, response: Response, /) -> Response:
        return response


class RequestParameter(Parameter):
    def resolve_response(self, response: Response, /) -> httpx.Request:
        return response.request

    def resolve_request(self, request: PreRequest, /) -> PreRequest:
        return request


class StatusCodeParameter(Parameter):
    def resolve_response(self, response: Response, /) -> int:
        return response.status_code


class StateParameter(
    ComposableSingletonParameter[str, Any], ResolvableSingletonParameter[str, Any]
):
    def parse_key(self, key: str, /) -> str:
        return key

    def parse_value(self, value: Any, /) -> Any:
        return value

    def build_consumer(self, key: str, value: Any) -> RequestConsumer:
        return StateConsumer(key, value).consume_request

    def build_request_resolver(self, key: str) -> RequestResolver[Any]:
        return StateResolver(key).resolve_request

    def build_response_resolver(self, key: str) -> ResponseResolver[Any]:
        return StateResolver(key).resolve_response


class ReasonParameter(Parameter):
    def resolve_response(self, response: Response, /) -> str:
        return response.reason_phrase



import dataclasses
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    MutableSequence,
    Optional,
    Sequence,
    TypeVar,
)

import httpx
from httpx import URL, BaseTransport, Cookies, Headers, Limits, QueryParams, Timeout
from httpx._auth import Auth
from mediate.protocols import MiddlewareCallable
from roster import Record
from typing_extensions import ParamSpec

from . import converters
from .composition import get_fields, validate_fields
from .constants import USER_AGENT
from .defaults import (
    DEFAULT_AUTH,
    DEFAULT_BASE_URL,
    DEFAULT_COOKIES,
    DEFAULT_ENCODING,
    DEFAULT_EVENT_HOOKS,
    DEFAULT_FOLLOW_REDIRECTS,
    DEFAULT_HEADERS,
    DEFAULT_LIMITS,
    DEFAULT_MAX_REDIRECTS,
    DEFAULT_PARAMS,
    DEFAULT_TIMEOUT,
    DEFAULT_TRUST_ENV,
)
from .enums import HTTPHeader, HTTPMethod
from .middlewares import Middleware
from .models import ClientOptions, PreRequest, Request, Response
from .operation import Operation, get_operation
from .types import (
    AuthTypes,
    CertTypes,
    CookiesTypes,
    DefaultEncodingTypes,
    EventHooks,
    HeadersTypes,
    ProxiesTypes,
    QueriesTypes,
    TimeoutTypes,
    URLTypes,
    VerifyTypes,
)
from .typing import Dependency

__all__: Sequence[str] = (
    "Session",
    "Client",
    "NeoClient",
)


T = TypeVar("T")

PS = ParamSpec("PS")
RT = TypeVar("RT")

D = TypeVar("D", bound=Dependency)


class BaseService:
    def __repr__(self) -> str:
        return f"{type(self).__name__}()"


@dataclass(init=False)
class Session(httpx.Client):
    auth: Optional[Auth]
    params: QueryParams
    headers: Headers
    cookies: Cookies
    timeout: Timeout
    follow_redirects: bool
    max_redirects: int
    event_hooks: Dict[str, List[Callable]]
    base_url: URL
    trust_env: bool

    def __init__(
        self,
        base_url: URLTypes = DEFAULT_BASE_URL,
        *,
        auth: Optional[AuthTypes] = DEFAULT_AUTH,
        params: Optional[QueriesTypes] = DEFAULT_PARAMS,
        headers: Optional[HeadersTypes] = DEFAULT_HEADERS,
        cookies: Optional[CookiesTypes] = DEFAULT_COOKIES,
        verify: VerifyTypes = True,
        cert: Optional[CertTypes] = None,
        http1: bool = True,
        http2: bool = False,
        proxies: Optional[ProxiesTypes] = None,
        mounts: Optional[Mapping[str, BaseTransport]] = None,
        timeout: TimeoutTypes = DEFAULT_TIMEOUT,
        follow_redirects: bool = DEFAULT_FOLLOW_REDIRECTS,
        limits: Limits = DEFAULT_LIMITS,
        max_redirects: int = DEFAULT_MAX_REDIRECTS,
        event_hooks: Optional[EventHooks] = DEFAULT_EVENT_HOOKS,
        transport: Optional[BaseTransport] = None,
        app: Optional[Callable[..., Any]] = None,
        trust_env: bool = DEFAULT_TRUST_ENV,
        default_encoding: DefaultEncodingTypes = DEFAULT_ENCODING,
    ) -> None:
        params = (
            converters.convert_query_params(params)
            if params is not None
            else QueryParams()
        )
        headers = (
            converters.convert_headers(headers) if headers is not None else Headers()
        )
        cookies = (
            converters.convert_cookies(cookies) if cookies is not None else Cookies()
        )
        timeout = (
            converters.convert_timeout(timeout) if timeout is not None else Timeout()
        )
        base_url = URL(base_url)

        # Set a default User-Agent header
        headers.setdefault(HTTPHeader.USER_AGENT, USER_AGENT)

        super().__init__(
            auth=auth,
            params=params,
            headers=headers,
            cookies=cookies,
            verify=verify,
            cert=cert,
            http1=http1,
            http2=http2,
            proxies=proxies,
            mounts=mounts,
            timeout=timeout,
            follow_redirects=follow_redirects,
            limits=limits,
            max_redirects=max_redirects,
            event_hooks=event_hooks,
            transport=transport,
            app=app,
            base_url=base_url,
            trust_env=trust_env,
            default_encoding=default_encoding,
        )


@dataclass(init=False)
class Client:
    client: Optional[httpx.Client]
    middleware: Middleware
    default_response: Optional[Dependency] = None
    request_dependencies: MutableSequence[Dependency] = field(default_factory=list)
    response_dependencies: MutableSequence[Dependency] = field(default_factory=list)

    def __init__(
        self,
        client: Optional[httpx.Client] = None,
        middleware: Optional[Middleware] = None,
        default_response: Optional[Dependency] = None,
        request_dependencies: Optional[Sequence[Dependency]] = None,
        response_dependencies: Optional[Sequence[Dependency]] = None,
    ) -> None:
        self.client = client
        self.middleware = middleware if middleware is not None else Middleware()
        self.default_response = default_response
        self.request_dependencies = (
            [*request_dependencies] if request_dependencies is not None else []
        )
        self.response_dependencies = (
            [*response_dependencies] if response_dependencies is not None else []
        )

    def bind(self, func: Callable[PS, RT], /) -> Callable[PS, RT]:
        operation: Operation = get_operation(func)

        # Create a clone of the operation's middleware, so that when the client's
        # middleware gets added, it doesn't mutate the original
        middleware: Middleware = Middleware()
        middleware.add_all(operation.middleware.record)

        # Create a clone of the operation's request dependencies, so that when
        # the client's request dependencies get added, it doesn't mutate the original
        request_dependencies: MutableSequence[Dependency] = []
        request_dependencies.extend(operation.request_dependencies)

        # Create a clone of the operation's response dependencies, so that when
        # the client's response dependencies get added, it doesn't mutate the original
        response_dependencies: MutableSequence[Dependency] = []
        response_dependencies.extend(operation.response_dependencies)

        bound_operation: Operation[PS, RT] = dataclasses.replace(
            operation,
            client=self.client,
            middleware=middleware,
            request_dependencies=request_dependencies,
            response_dependencies=response_dependencies,
        )

        # If the operation doesn't have a response, use the client's default response
        # if one is available
        if bound_operation.response is None and self.default_response is not None:
            bound_operation.response = self.default_response

        # Add the client's middleware
        bound_operation.middleware.add_all(self.middleware.record)

        # Add the client's request dependencies
        request_dependencies.extend(self.request_dependencies)

        # Add the client's response dependencies
        response_dependencies.extend(self.response_dependencies)

        return bound_operation.wrapper

    def request(
        self,
        method: str,
        endpoint: str,
        /,
        *,
        response: Optional[Dependency] = None,
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        client_options: ClientOptions = ClientOptions()
        request_options: PreRequest = PreRequest(
            method=method,
            url=endpoint,
        )

        operation_response: Optional[Callable] = None

        if response is not None:
            operation_response = response
        elif self.default_response is not None:
            operation_response = self.default_response

        def decorator(func: Callable[PS, RT], /) -> Callable[PS, RT]:
            # Create a copy of the client's middleware
            middleware: Middleware = Middleware()
            middleware.add_all(self.middleware.record)

            # Create a copy of the client's request dependencies
            request_dependencies: MutableSequence[Dependency] = []
            request_dependencies.extend(self.request_dependencies)

            # Create a copy of the client's response dependencies
            response_dependencies: MutableSequence[Dependency] = []
            response_dependencies.extend(self.response_dependencies)

            operation: Operation[PS, RT] = Operation(
                func=func,
                client_options=client_options,
                request_options=request_options,
                client=self.client,
                middleware=middleware,
                response=operation_response,
                request_dependencies=request_dependencies,
                response_dependencies=response_dependencies,
            )

            # Validate operation function parameters are acceptable
            validate_fields(get_fields(operation.request_options, func))

            return operation.wrapper

        return decorator

    def put(
        self, endpoint: str, /, *, response: Optional[Dependency] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HTTPMethod.PUT.name, endpoint, response=response)

    def get(
        self, endpoint: str, /, *, response: Optional[Dependency] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HTTPMethod.GET.name, endpoint, response=response)

    def post(
        self, endpoint: str, /, *, response: Optional[Dependency] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HTTPMethod.POST.name, endpoint, response=response)

    def head(
        self, endpoint: str, /, *, response: Optional[Dependency] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HTTPMethod.HEAD.name, endpoint, response=response)

    def patch(
        self, endpoint: str, /, *, response: Optional[Dependency] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HTTPMethod.PATCH.name, endpoint, response=response)

    def delete(
        self, endpoint: str, /, *, response: Optional[Dependency] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HTTPMethod.DELETE.name, endpoint, response=response)

    def options(
        self, endpoint: str, /, *, response: Optional[Dependency] = None
    ) -> Callable[[Callable[PS, RT]], Callable[PS, RT]]:
        return self.request(HTTPMethod.OPTIONS.name, endpoint, response=response)

    def request_depends(self, dependency: D, /) -> D:
        self.request_dependencies.append(dependency)

        return dependency

    def response_depends(self, dependency: D, /) -> D:
        self.response_dependencies.append(dependency)

        return dependency


class NeoClient(Client):
    def __init__(
        self,
        base_url: URLTypes = DEFAULT_BASE_URL,
        *,
        auth: Optional[AuthTypes] = DEFAULT_AUTH,
        params: Optional[QueriesTypes] = DEFAULT_PARAMS,
        headers: Optional[HeadersTypes] = DEFAULT_HEADERS,
        cookies: Optional[CookiesTypes] = DEFAULT_COOKIES,
        verify: VerifyTypes = True,
        cert: Optional[CertTypes] = None,
        http1: bool = True,
        http2: bool = False,
        proxies: Optional[ProxiesTypes] = None,
        mounts: Optional[Mapping[str, BaseTransport]] = None,
        timeout: TimeoutTypes = DEFAULT_TIMEOUT,
        follow_redirects: bool = DEFAULT_FOLLOW_REDIRECTS,
        limits: Limits = DEFAULT_LIMITS,
        max_redirects: int = DEFAULT_MAX_REDIRECTS,
        event_hooks: Optional[EventHooks] = DEFAULT_EVENT_HOOKS,
        transport: Optional[BaseTransport] = None,
        app: Optional[Callable[..., Any]] = None,
        trust_env: bool = DEFAULT_TRUST_ENV,
        default_encoding: DefaultEncodingTypes = DEFAULT_ENCODING,
        middleware: Optional[Sequence[MiddlewareCallable[Request, Response]]] = None,
        default_response: Optional[Dependency] = None,
        request_dependencies: Optional[Sequence[Dependency]] = None,
        response_dependencies: Optional[Sequence[Dependency]] = None,
    ) -> None:
        super().__init__(
            client=Session(
                auth=auth,
                params=params,
                headers=headers,
                cookies=cookies,
                verify=verify,
                cert=cert,
                http1=http1,
                http2=http2,
                proxies=proxies,
                mounts=mounts,
                timeout=timeout,
                follow_redirects=follow_redirects,
                limits=limits,
                max_redirects=max_redirects,
                event_hooks=event_hooks,
                transport=transport,
                app=app,
                base_url=base_url,
                trust_env=trust_env,
                default_encoding=default_encoding,
            ),
            middleware=(
                Middleware(record=Record(middleware))
                if middleware is not None
                else Middleware()
            ),
            default_response=default_response,
            request_dependencies=(
                request_dependencies if request_dependencies is not None else []
            ),
            response_dependencies=(
                response_dependencies if response_dependencies is not None else []
            ),
        )



import inspect
from typing import Any, Callable, Dict, MutableSequence, Optional, Sequence, Tuple, Type

from annotate.utils import has_annotation
from mediate.protocols import MiddlewareCallable

from .client import Client
from .enums import Entity
from .errors import ServiceInitialisationError
from .middlewares import Middleware
from .models import Request, Response
from .operation import Operation, get_operation, has_operation
from .specification import ClientSpecification
from .typing import Dependency

__all__: Sequence[str] = ("Service",)


class ServiceMeta(type):
    _spec: ClientSpecification

    def __new__(
        mcs: Type["ServiceMeta"], name: str, bases: Tuple[type], attrs: Dict[str, Any]
    ) -> type:
        def __init__(self) -> None:
            service_middleware: Sequence[MiddlewareCallable[Request, Response]] = [
                member
                for _, member in inspect.getmembers(self)
                if has_annotation(member, Entity.MIDDLEWARE)
            ]
            service_responses: Sequence[Dependency] = [
                member
                for _, member in inspect.getmembers(self)
                if has_annotation(member, Entity.RESPONSE)
            ]
            service_request_dependencies: Sequence[Dependency] = [
                member
                for _, member in inspect.getmembers(self)
                if has_annotation(member, Entity.REQUEST_DEPENDENCY)
            ]
            service_response_dependencies: Sequence[Dependency] = [
                member
                for _, member in inspect.getmembers(self)
                if has_annotation(member, Entity.RESPONSE_DEPENDENCY)
            ]

            if len(service_responses) > 1:
                raise ServiceInitialisationError(
                    f"Found {len(service_responses)} service responses, expected at most 1"
                )

            middleware: Middleware = Middleware()

            middleware.add_all(self._spec.middleware.record)
            middleware.add_all(service_middleware)

            response: Optional[Dependency] = self._spec.default_response

            # If a service-level response was defined, use this instead of the one
            # defined within the specification
            if service_responses:
                response = service_responses[0]

            request_dependencies: MutableSequence[Dependency] = []

            request_dependencies.extend(self._spec.request_dependencies)
            request_dependencies.extend(service_request_dependencies)

            response_dependencies: MutableSequence[Dependency] = []

            response_dependencies.extend(self._spec.response_dependencies)
            response_dependencies.extend(service_response_dependencies)

            self._client = Client(
                client=self._spec.options.build(),
                middleware=middleware,
                default_response=response,
                request_dependencies=request_dependencies,
                response_dependencies=response_dependencies,
            )

            for member_name, member in inspect.getmembers(self):
                if not has_operation(member):
                    continue

                bound_operation_func: Callable = self._client.bind(member)
                bound_operation_method: Callable = bound_operation_func.__get__(self)

                bound_operation: Operation = get_operation(bound_operation_method)

                bound_operation.func = bound_operation_method

                setattr(self, member_name, bound_operation_method)

        attrs["_spec"] = ClientSpecification()
        attrs["__init__"] = __init__

        typ: type = super().__new__(mcs, name, bases, attrs)

        return typ


class Service(metaclass=ServiceMeta):
    _client: Client

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"
